@@ -34,6 +34,25 @@ class JittableModuleTest(unittest.TestCase):
     assert isinstance(JittableMoreAwesomeModel, EvenMoreAwesomeModel)
     assert not isinstance(JittableMoreAwesomeModel, MyAwesomeModel)
 
+  def test_functional_call_callable(self):
+
+    def outer_function(model, x):
+      return x + 1
+
+    model = MyAwesomeModel()
+    jittable_module = interop.JittableModule(model)
+
+    # Check if the jittable module can be called like a function
+    input_tensor = torch.randn(1, 3, 224, 224)
+    expected_output = input_tensor + 1
+
+    output = jittable_module.functional_call(outer_function,
+                                             jittable_module.params,
+                                             jittable_module.buffers,
+                                             input_tensor)
+
+    assert torch.equal(output, expected_output)
+
 
 if __name__ == '__main__':
   unittest.main()
